using System;
using System.Runtime.InteropServices;
using System.ComponentModel;
using System.Threading;

namespace Win32
{
	internal class HiPerfTimer
	{
		[DllImport("Kernel32.dll")]
		private static extern bool QueryPerformanceCounter(
			out long lpPerformanceCount);

		[DllImport("Kernel32.dll")]
		private static extern bool QueryPerformanceFrequency(
			out long lpFrequency);

		private long startTime, stopTime;
		private long freq;

		// Constructor
		public HiPerfTimer()
		{
			startTime = 0;
			stopTime  = 0;

			if (QueryPerformanceFrequency(out freq) == false)
			{
				// high-performance counter not supported
				throw new Win32Exception();
			}
		}

		/// <summary>
		/// Start the timer
		/// </summary>
		public void Start()
		{
			// lets do the waiting threads their work
			Thread.Sleep(0);

			QueryPerformanceCounter(out startTime);
		}

		/// <summary>
		/// Stop the timer
		/// </summary>
		public void Stop()
		{
			QueryPerformanceCounter(out stopTime);
		}

		/// <summary>
		/// Returns the duration of the timer (in seconds) 
		/// </summary>
		public double Duration
		{
			get
			{
				return (double)(stopTime - startTime) / (double) freq;
			}
		}
		
		/// <summary>
		/// Returns the duration of the timer (in seconds) while process is running
		/// </summary>
		public double CurrentDuration
		{
			get
			{
				long time=0;
				QueryPerformanceCounter(out time);
				return (double)(time - startTime) / (double) freq;
			}
		}
	}
}