//
// Objects classes declaring:
//   Point3f - Floating precision 3D point 
//   Point3d - Double precision 3D point 
//   cTriangle - Triangle

using System;

namespace Raytracer
{
	/// <summary>
	/// Double precision coordinate
	/// </summary>
	/// <summary>
	/// Double point precision coordinate
	/// </summary>
	public class Point3d 
	{
		private double[] v;

		/// <summary>
		/// X coordinate
		/// </summary>
		public double x 
		{
			get { return v[0]; }
			set { v[0] = value; }
		}

		/// <summary>
		/// Y coordinate
		/// </summary>
		public double y 
		{
			get { return v[1]; }
			set { v[1] = value; }
		}

		/// <summary>
		/// Z coordinate
		/// </summary>
		public double z 
		{
			get { return v[2]; } 
			set { v[2] = value; }
		}

		/// <summary>
		/// Converts the coordinate to a string
		/// </summary>
		/// <returns>Returns a string representation of the coordinate using the format: x , y , z</returns>
		public override string ToString()
		{
			return x.ToString() + " , " + y.ToString() + " , " + z.ToString();
		}

		/// <summary>
		/// Initializes a new instance of a double precision 3D point.
		/// </summary>
		public Point3d() 
		{
			v=new double[3];
			v[0]=0;v[1]=0;v[2]=0;
		}

		/// <summary>
		/// Initializes a new instance of a double precision 3D point
		/// </summary>
		/// <param name="X">X coordinate</param>
		/// <param name="Y">Y coordinate</param>
		/// <param name="Z">Z coordinate</param>
		public Point3d(double X,double Y,double Z)
		{
			v=new double[3];
			v[0]=X;v[1]=Y;v[2]=Z;
		}

		/// <summary>
		/// Returns part of coordinate. Index 0 = X, Index 1 = Y, Index 2 = Z
		/// </summary>
		public double this[uint index]
		{
			get 
			{
				if(index>2)
					throw(new System.Exception("Vector index out of bounds"));
				return v[index];
			}
			set 
			{
				if(index>2)
					throw(new System.Exception("Vector index out of bounds"));
				v[index] = value;
			}
		}

		/// <summary>
		/// Vector + Vector
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns></returns>
		public static Point3d operator +(Point3d v1,Point3d v2)
		{ return new Point3d(v1.x+v2.x,v1.y+v2.y,v1.z+v2.z); }

		/// <summary>
		/// Vector + Vector
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns></returns>
		public static Point3d operator +(Point3f v1,Point3d v2)
		{ return new Point3d(v1.x+v2.x,v1.y+v2.y,v1.z+v2.z); }
		
		/// <summary>
		/// Vector + Vector
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns></returns>
		public static Point3d operator +(Point3d v1,Point3f v2)
		{ return new Point3d(v1.x+v2.x,v1.y+v2.y,v1.z+v2.z); }

		/// <summary>
		/// Vector * Scalar
		/// </summary>
		/// <param name="m">Vector</param>
		/// <param name="i">Scalar (int)</param>
		/// <returns></returns>
		public static Point3d operator *(Point3d m,int i)
		{ return new Point3d(m.x*i,m.y*i,m.z*i); }
		
		/// <summary>
		/// Vector * Scalar
		/// </summary>
		/// <param name="m">Vector</param>
		/// <param name="d">Scalar (double)</param>
		/// <returns></returns>
		public static Point3d operator *(Point3d m,double d)
		{ return new Point3d(m.x*d,m.y*d,m.z*d); }
		
		/// <summary>
		/// Vector * Vector (Cross product)
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns>Cross product</returns>
		public static Point3d operator *(Point3d v1,Point3d v2)
		{ return new Point3d(v1.y*v2.z - v1.z*v2.y, v1.z*v2.x - v1.x*v2.z, v1.x*v2.y - v1.y*v2.x); }
		
		/// <summary>
		/// Vector - Vector
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns>Cross product</returns>
		public static Point3d operator -(Point3d v1,Point3d v2)
		{ return new Point3d(v1.x-v2.x,v1.y-v2.y,v1.z-v2.z); }
	}

	/// <summary>
	/// Floating point precision coordinate
	/// </summary>
	public class Point3f 
	{
		private float[] v;
	
		/// <summary>
		/// Initializes a new instance of a floating precision 3D point.
		/// </summary>		
		public Point3f() 
		{
			v=new float[3];
			v[0]=0;v[1]=0;v[2]=0;
		}

		/// <summary>
		/// Initializes a new instance of a floating precision 3D point
		/// </summary>
		/// <param name="X">X coordinate</param>
		/// <param name="Y">Y coordinate</param>
		/// <param name="Z">Z coordinate</param>
		public Point3f(float X,float Y,float Z)
		{
			v=new float[3];
			v[0]=X;v[1]=Y;v[2]=Z;
		}

		/// <summary>
		/// X coordinate
		/// </summary>
		public float x 
		{
			get { return v[0]; }
			set { v[0] = value; }
		}

		/// <summary>
		/// Y coordinate
		/// </summary>
		public float y 
		{
			get { return v[1]; }
			set { v[1] = value; }
		}

		/// <summary>
		/// Z coordinate
		/// </summary>
		public float z 
		{
			get { return v[2]; } 
			set { v[2] = value; }
		}

		/// <summary>
		/// Converts the coordinate to a string
		/// </summary>
		/// <returns>Returns a string representation of the coordinate using the format: x , y , z</returns>
		public override string ToString()
		{
			return x.ToString() + " , " + y.ToString() + " , " + z.ToString();
		}

		/// <summary>
		/// Returns part of coordinate. Index 0 = X, Index 1 = Y, Index 2 = Z
		/// </summary>
		public float this[uint index]
		{
			get 
			{
				if(index>2)
					throw(new System.Exception("Vector index out of bounds"));
				return v[index];
			}
			set 
			{
				if(index>2)
					throw(new System.Exception("Vector index out of bounds"));
				v[index] = value;
			}
		}

		/// <summary>
		/// Vector + Vector
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns></returns>
		public static Point3f operator +(Point3f v1,Point3f v2)
		{ return new Point3f(v1.x+v2.x,v1.y+v2.y,v1.z+v2.z); }
		
		/// <summary>
		/// Vector * Scalar
		/// </summary>
		/// <param name="m">Vector</param>
		/// <param name="i">Scalar (int)</param>
		/// <returns></returns>
		public static Point3f operator *(Point3f m,int i)
		{ return new Point3f(m.x*i,m.y*i,m.z*i); }
		
		/// <summary>
		/// Vector * Scalar
		/// </summary>
		/// <param name="m">Vector</param>
		/// <param name="f">Scalar (float)</param>
		/// <returns></returns>
		public static Point3f operator *(Point3f m,float f)
		{ return new Point3f(m.x*f,m.y*f,m.z*f); }
		
		/// <summary>
		/// Vector * Vector (Cross product)
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns>Cross product</returns>
		public static Point3f operator *(Point3f v1,Point3f v2)
		{ return new Point3f(v1.y*v2.z - v1.z*v2.y, v1.z*v2.x - v1.x*v2.z, v1.x*v2.y - v1.y*v2.x); }
		
		/// <summary>
		/// Vector - Vector
		/// </summary>
		/// <param name="v1">Vector</param>
		/// <param name="v2">Vector</param>
		/// <returns>Cross product</returns>
		public static Point3f operator -(Point3f v1,Point3f v2)
		{ return new Point3f(v1.x-v2.x,v1.y-v2.y,v1.z-v2.z); }

		/// <summary>
		/// Returns the maximum values of the two vectors
		/// </summary>
		/// <param name="aabb2"></param>
		/// <returns></returns>
		public Point3f Max(Point3f aabb2)
		{
			return new Point3f((this.x>aabb2.x?this.x:aabb2.x),
				(this.y>aabb2.y?this.y:aabb2.y),
				(this.z>aabb2.z?this.z:aabb2.z));
		}

		/// <summary>
		/// Returns the minimum values of the two vectors
		/// </summary>
		/// <param name="aabb2"></param>
		/// <returns></returns>
		public Point3f Min(Point3f aabb2)
		{
			return new Point3f((this.x<aabb2.x?this.x:aabb2.x),
				(this.y<aabb2.y?this.y:aabb2.y),
				(this.z<aabb2.z?this.z:aabb2.z));
		}

		/// <summary>
		/// Returns "volume" of vector
		/// </summary>
		public float Prod
		{
			get
			{
				return this.x*this.y*this.z;
			}
		}
	}

	/// <summary>
	/// Triangle class
	/// </summary>
	public class cTriangle 
	{
		private Point3f[] v;

		/// <summary>
		/// Initializes a new triangle
		/// </summary>
		public cTriangle()
		{
			v=new Point3f[3];
			v[0] = new Point3f();
			v[1] = new Point3f();
			v[2] = new Point3f();
		}

		/// <summary>
		/// Initializes a new triangle from tree vertices
		/// </summary>
		/// <param name="v1">Vertex 1</param>
		/// <param name="v2">Vertex 2</param>
		/// <param name="v3">Vertex 3</param>
		public cTriangle(Point3f v1,Point3f v2,Point3f v3)
		{
			v=new Point3f[3];
			v[0] = v1;
			v[1] = v2;
			v[2] = v3;
		}

		/// <summary>
		/// Returns midpoint of triangle
		/// </summary>
		public Point3f MidPoint
		{
			get 
			{
				return (v[0]+v[1]+v[2])*(1/3f);
			}
		}

		/// <summary>
		/// Returns n'th vertex
		/// </summary>
		public Point3f this[uint index]
		{
			get 
			{
				if(index>2)
					throw(new System.Exception("vertex index out of bounds"));
				return v[index];
			}
			set 
			{
				if(index>2)
					throw(new System.Exception("vertex index out of bounds"));
				v[index] = value;
			}
		}

		/// <summary>
		/// Get the AABB for this triangle (used for making group aabb)
		/// </summary>
		public AABB TriAABB
		{
			get 
			{
				// new aabb with min as minimum of 3 verts and maximum max of 3 verts
				return new AABB(v[0].Min(v[1]).Min(v[2]),
					v[0].Max(v[1]).Max(v[2]));
			}
		}
	}
}
